@@ -115,10 +115,12 @@ import traceback
 from . import js_context
 from .hex_codec import hex_to_bytearray, bytearray_to_hex
 from pprint import pprint
+import numpy as np
+from jupyter_ui_poll import run_ui_poll_loop
 
 # In the IPython context get_ipython is a builtin.
 # get a reference to the IPython notebook object.
-ip = IPython.get_ipython()
+#ip = IPython.get_ipython()   # not used
 
 # For creating unique DOM identities
 IDENTITY_COUNTER = [int(time.time() * 100) % 10000000]
@@ -143,9 +145,18 @@ FRAGILE_JS_REFERENCE = "_FRAGILE_JS_REFERENCE"
 # Reference used for method resolution
 FRAGILE_THIS = "_FRAGILE_THIS"
 
+SEND_FRAGILE_JS_REFERENCE = "_SEND_FRAGILE_JS_REFERENCE"
+
 # Message segmentation size default
 BIG_SEGMENT = 1000000
 
+class SyncTimeOutError(RuntimeError):
+    "The sync operation between the kernel and Javascript timed out."
+
+class JavascriptException(ValueError):
+    "The sync operation caused an exception in the Javascript interpreter."
+
+
 @widgets.register
 class JSProxyWidget(widgets.DOMWidget):
     """Introspective javascript proxy widget.""" 
@@ -153,8 +164,8 @@ class JSProxyWidget(widgets.DOMWidget):
     _model_name = Unicode('JSProxyModel').tag(sync=True)
     _view_module = Unicode('jp_proxy_widget').tag(sync=True)
     _model_module = Unicode('jp_proxy_widget').tag(sync=True)
-    _view_module_version = Unicode('^1.0.5').tag(sync=True)
-    _model_module_version = Unicode('^1.0.5').tag(sync=True)
+    _view_module_version = Unicode('^1.0.7').tag(sync=True)
+    _model_module_version = Unicode('^1.0.7').tag(sync=True)
 
     # traitlet port to use for sending commands to javascript
     #commands = traitlets.List([], sync=True)
@@ -191,7 +202,7 @@ class JSProxyWidget(widgets.DOMWidget):
         ##pr "registered on_msg(handle_custom_message)"
         self.on_msg(self.handle_custom_message_wrapper)
         self.buffered_commands = []
-        self.commands_awaiting_render = []
+        #self.commands_awaiting_render = []
         self.last_commands_sent = []
         self.last_callback_results = None
         self.results = []
@@ -204,8 +215,22 @@ class JSProxyWidget(widgets.DOMWidget):
         self.js_init("""
             // make the window accessible through the element
             element.window = window;
+
+            // Initialize caching slots.
             element._FRAGILE_THIS = null;
-        """)
+            element._FRAGILE_JS_REFERENCE = null;
+
+            // The following is used for sending synchronous values.
+            element._SEND_FRAGILE_JS_REFERENCE = function(ms_delay) {
+                // xxxxx add a delay to allow the Python side to be ready for the response (???)
+                ms_delay = ms_delay || 100;
+                var ref = element._FRAGILE_JS_REFERENCE;
+                var delayed = function () {
+                    RECEIVE_FRAGILE_REFERENCE(ref);
+                };
+                setTimeout(delayed, ms_delay);
+            };
+        """, RECEIVE_FRAGILE_REFERENCE=self._RECEIVE_FRAGILE_REFERENCE)
 
     def set_element(self, slot_name, value):
         """
@@ -287,6 +312,15 @@ class JSProxyWidget(widgets.DOMWidget):
         "Convenience access to window.setTimeout in Javascript"
         self.element.window.setTimeout(callable, milliseconds)
 
+    def on_rendered(self, callable, *positional, **keyword):
+        """
+        After the widget has rendered, call the callable using the provided arguments.
+        This can be used to initialize an animation after the widget is visible, for example.
+        """
+        def call_it():
+            return callable(*positional, **keyword)
+        self.js_init("call_it();", call_it=call_it)
+
     def handle_error_msg(self, att_name, old, new):
         if self.print_on_error:
             print("new error message: " + new)
@@ -294,8 +328,11 @@ class JSProxyWidget(widgets.DOMWidget):
     def handle_rendered(self, att_name, old, new):
         "when rendered send any commands awaiting the render event."
         try:
-            if self.commands_awaiting_render:
-                self.send_commands([])
+            #if self.commands_awaiting_render:
+                #self.send_commands([])
+            if self.auto_flush:
+                #("xxxx flushing on render")
+                self.flush()
             self.status= "Rendered."
         except Exception as e:
             self.error_msg = repr(e)
@@ -388,21 +425,37 @@ class JSProxyWidget(widgets.DOMWidget):
         return prefix + str(IDENTITY_COUNTER[0])
 
     def __call__(self, command):
+        "Send command convenience."
+        return self.buffer_command(command)
+
+    def buffer_command(self, command):
         "Add a command to the buffered commands. Convenience."
-        #self.buffered_commands.append(command)
-        self.send_command(command)
+        self.buffer_commands([command])
+        return command
+
+    def buffer_commands(self, commands):
+        self.buffered_commands.extend(commands)
         if self.auto_flush:
             self.flush()
-        return command
+        return commands
 
     def seg_flush(self, results_callback=None, level=1, segmented=BIG_SEGMENT):
         "flush a potentially large command sequence, segmented."
         return self.flush(results_callback, level, segmented)
 
+    error_on_flush = False  # Primarily for debugging
+
     def flush(self, results_callback=None, level=1, segmented=None):
         "send the buffered commands and clear the buffer. Convenience."
+        if not self.rendered:
+            #("XXXX not flushing before render", len(self.buffered_commands))
+            self.status = "deferring flush until render"
+            return None
+        if self.error_on_flush:
+            raise ValueError("flush is disabled")
         commands = self.buffered_commands
         self.buffered_commands = []
+        #("XXXXX now flushing", len(commands))
         result = self.send_commands(commands, results_callback, level, segmented=segmented)
         self._send_counter += 1
         return result
@@ -438,6 +491,31 @@ class JSProxyWidget(widgets.DOMWidget):
         if onsuccess:
             onsuccess()
 
+    def in_dialog(
+            self,
+            title="",
+            autoOpen=True,
+            buttons=None,  # dict of label to callback
+            height="auto",
+            width=300,
+            modal=False,
+            **other_options,
+        ):
+        """
+        Pop the widget into a floating jQueryUI dialog. See https://api.jqueryui.com/1.9/dialog
+        """
+        self.check_jquery()
+        options = clean_dict(
+            title=title,
+            autoOpen=autoOpen,
+            buttons=buttons,
+            height=height,
+            width=width,
+            modal=modal,
+            **other_options,
+            )
+        self.element.dialog(options)
+
     _require_checked = False
     _needs_requirejs = False
     _delayed_require_actions = None
@@ -618,10 +696,13 @@ class JSProxyWidget(widgets.DOMWidget):
         if check:
             debug_check_commands(commands)
         if self.rendered:
-            # also send any commands awaiting the render event.
-            if self.commands_awaiting_render:
-                commands = commands + self.commands_awaiting_render
-                self.commands_awaiting_render = None
+            # also send buffered commands
+            #if self.commands_awaiting_render:
+            #    commands = commands + self.commands_awaiting_render
+            #    self.commands_awaiting_render = None
+            if self.buffered_commands:
+                commands = self.buffered_commands + commands
+                self.buffered_commands = []
             payload = [count, commands, level]
             if results_callback is not None:
                 self.identifier_to_callback[count] = results_callback
@@ -636,7 +717,8 @@ class JSProxyWidget(widgets.DOMWidget):
         else:
             # wait for render event before sending commands.
             ##pr "waiting for render!", commands
-            self.commands_awaiting_render.extend(commands)
+            #self.commands_awaiting_render.extend(commands)
+            self.buffered_commands.extend(commands)
             return ("awaiting render", commands)
 
     def send_segmented_message(self, frag_ind, final_ind, payload, segmented):
@@ -653,7 +735,70 @@ class JSProxyWidget(widgets.DOMWidget):
             cursor = next_cursor
         json_tail = json_str[cursor:]
         self.send_custom_message(final_ind, json_tail)
+
+    _synced_command_result = None
+    _synced_command_evaluated = False
+    _synced_command_timed_out = False
+    _synced_command_timeout_time = None
+
+    def evaluate(self, command, level=3, timeout=3000, ms_delay=100):
+        "Evaluate the command and return the converted javascript value."
+        # temporarily disable error prints
+        print_on_error = self.print_on_error
+        old_err = self.error_msg
+        try:
+            # Note: if the command buffer has not been flushed other operations may set the error_msg
+            self.print_on_error = False
+            self.error_msg = ""
+            self._synced_command_result = None
+            self._synced_command_timed_out = False
+            self._synced_command_evaluated = False
+            self._synced_command_timeout_time = None
+            start = time.time()
+            if timeout is not None and timeout > 0:
+                self._synced_command_timeout_time = start + timeout
+            start = self._synced_command_start_time = time.time()
+            self._send_synced_command(command, level, ms_delay=ms_delay)
+            run_ui_poll_loop(self._sync_complete)
+            if self._synced_command_timed_out:
+                raise TimeoutError("wait: %s, started: %s; gave up %s" % (timeout, start, time.time()))
+            assert self._synced_command_evaluated, repr((self._synced_command_evaluated, self._synced_command_result))
+            error_msg = self.error_msg
+            result = self._synced_command_result
+            if error_msg:
+                if error_msg == result:
+                    raise JavascriptException("sync error: " + repr(error_msg))
+                else:
+                    old_err = error_msg
+            return result
+        finally:
+            # restore error prints if formerly enabled
+            self.error_msg = old_err
+            self.print_on_error = print_on_error
+
+    def _send_synced_command(self, command, level, ms_delay=100):
+        if self.last_fragile_reference is not command:
+            set_ref = SetMaker(self.get_element(), FRAGILE_JS_REFERENCE, command)
+            self.buffer_command(set_ref)
+        #self.element._SEND_FRAGILE_JS_REFERENCE()
+        get_ref = CallMaker("method", self.get_element(), SEND_FRAGILE_JS_REFERENCE, ms_delay)
+        self.buffer_command(get_ref)
+        self.flush()
+
+    def _RECEIVE_FRAGILE_REFERENCE(self, value):
+        self._synced_command_result = value
+        self._synced_command_evaluated = True
+
+    def _sync_complete(self):
+        if self._synced_command_timeout_time is not None:
+            self._synced_command_timed_out = (time.time() > self._synced_command_timeout_time)
+        test = self._synced_command_evaluated or self._synced_command_timed_out
+        if test:
+            return test
+        else:
+            return None  # only None signals end of polling loop.
     
+    """ # doesn't work: not used.
     def evaluate(self, command, level=1, timeout=3000):
         "Send one command and wait for result.  Return result."
         results = self.evaluate_commands([command], level, timeout)
@@ -677,6 +822,7 @@ class JSProxyWidget(widgets.DOMWidget):
                 raise Exception("Timeout waiting for command results: " + repr(timeout))
             ip.kernel.do_one_iteration()
         return result_list[0]
+    """
 
     def seg_callback(self, callback_function, data, level=1, delay=False, segmented=BIG_SEGMENT):
         """
@@ -866,6 +1012,18 @@ class JSProxyWidget(widgets.DOMWidget):
         # Non-lists are untranslated (but should be JSON compatible).
         return command
 
+    def delay_flush(self):
+        """
+        Context manager to group a large number of operations into one message.
+        This can prevent flooding messages over the ZMQ communications link between
+        the Javascript front end and the Python kernel backend.
+
+        >>> with widget.delay_flush():
+        ...    many_operations(widget)
+        ...    even_more_operations(widget)
+        """
+        return DisableFlushContextManager(self)
+
 def indent_string(s, level, indent="    "):
     lindent = indent * level
     return s.replace("\n", "\n" + lindent)
@@ -894,6 +1052,30 @@ def to_javascript(thing, level=0, indent=None, comma=","):
     return result
 
 
+# Adapted from jp_doodle.dual_canvas.DisableRedrawContextManager
+
+class DisableFlushContextManager(object):
+    """
+    Temporarily disable flushes and also collect widget messages into a single group.
+    This can speed up widget interactions and prevent the communication channel from flooding.
+    """
+
+    def __init__(self, canvas):
+        self.canvas = canvas
+        self.save_flush = canvas.auto_flush
+
+    def __enter__(self):
+        canvas = self.canvas
+        self.save_flush = canvas.auto_flush
+        canvas.auto_flush = False
+
+    def __exit__(self, type, value, traceback):
+        canvas = self.canvas
+        canvas.auto_flush = self.save_flush
+        if (self.save_flush):
+            canvas.flush()
+
+
 class ElementWrapper(object):
 
     """
@@ -929,7 +1111,7 @@ class ElementWrapper(object):
         if isinstance(value, CommandMakerSuperClass):
             ref = value.reference()
         command = SetMaker(self.widget_element, name, ref)
-        self.widget.send_commands([command])
+        self.widget.buffer_commands([command])
         return LazyGet(self.widget, self.widget_element, name)
 
 class StaleFragileJavascriptReference(ValueError):
@@ -966,7 +1148,7 @@ class LazyCommandSuperClass(CommandMakerSuperClass):
         #return repr("Javascript disabled for lazy commands: " + repr(type(self)))
 
     def __getattr__(self, attribute):
-        if '_ipython' in attribute:
+        if type(attribute) is str and '_ipython' in attribute:
             return "attributes mentioning _ipython are forbidden because they cause infinite recursions: " + repr(attribute)
         return LazyGet(self.for_widget, self, attribute)
 
@@ -985,6 +1167,15 @@ class LazyCommandSuperClass(CommandMakerSuperClass):
             return f(*args)
         return result
 
+    def sync_value(self, timeout=3000, level=3, ms_delay=100):
+        """
+        Return the converted javascript-side value for this command.
+        """
+        for_widget = self.for_widget
+        for_widget.evaluate(self, timeout=timeout, level=level, ms_delay=ms_delay)
+        return for_widget._synced_command_result
+
+
 class LazyGet(LazyCommandSuperClass):
 
     def __init__(self, for_widget, for_target, attribute):
@@ -997,7 +1188,7 @@ class LazyGet(LazyCommandSuperClass):
         attr_ref = MethodMaker(MethodMaker(for_widget.get_element(), FRAGILE_THIS), attribute)
         set_ref = SetMaker(for_widget.get_element(), FRAGILE_JS_REFERENCE, attr_ref)
         # execute immediately on init
-        for_widget.send_commands([set_this, set_ref])
+        for_widget.buffer_commands([set_this, set_ref])
         # use fragile ref to find value, until the cached value is replaced
         for_widget.last_fragile_reference = self
 
@@ -1033,7 +1224,7 @@ class LazyCall(LazyCommandSuperClass):
             FRAGILE_JS_REFERENCE,
             CallMaker("function", self.for_target.reference(), *args)
         )
-        for_widget.send_commands([set_ref])
+        for_widget.buffer_commands([set_ref])
         for_widget.last_fragile_reference = self
 
     def _cmd(self):
@@ -1053,7 +1244,7 @@ class LazyMethodCall(LazyCommandSuperClass):
             FRAGILE_JS_REFERENCE,
             CallMaker("method", for_method.this_reference(), for_method.attribute, *args)
         )
-        for_widget.send_commands([set_ref])
+        for_widget.buffer_commands([set_ref])
         for_widget.last_fragile_reference = self
 
     def _cmd(self):
@@ -1301,7 +1492,6 @@ class InvalidCommand(Exception):
     "Invalid command"
 
 
-
 def debug_check_commands(command):
     "raise an error if the command is not a basic json structure"
     if command is None:
@@ -1318,3 +1508,23 @@ def debug_check_commands(command):
         return debug_check_commands(list(command.items()))
     # otherwise
     raise InvalidCommand(repr(command))
+
+
+
+def clean_dict(**kwargs):
+    "Like dict but with no None values and make some values JSON serializable."
+    # This function is generally useful for passing information to proxy widgets
+    result = {}
+    for kw in kwargs:
+        v = kwargs[kw]
+        if v is not None:
+            if isinstance(v, np.ndarray):
+                # listiffy arrays -- maybe should be done elsewhere
+                v = v.tolist()
+            if isinstance(v, np.floating):
+                v = float(v)
+            if type(v) is tuple:
+                v = list(v)
+            result[kw] = v
+    return result
+
